import hashlib
from functools import partial
import os

import keyring
from textual import on, work
from textual.app import App
from textual.binding import Binding

from lucidwallet.helpers import InitAppResponse, init_app
from lucidwallet.screens import (
    CreateWallet,
    EncryptionPassword,
    FirstRun,
    ImportFromMnemonic,
    MnemonicOverlay,
    WalletLanding,
)

import asyncio

import httpx
import importlib_metadata

package = "lucidwallet"


class FluxWallet(App[None]):
    CSS_PATH = "app.css"
    SCREENS = {
        "welcome": FirstRun(),
        "create_wallet": CreateWallet(),
        "from_mnemonic": ImportFromMnemonic(),
    }
    BINDINGS = [
        ("ctrl+t", "app.toggle_dark", "Toggle Dark mode"),
        Binding("ctrl+c,ctrl+q", "app.quit", "Quit", show=True),
    ]

    async def new_version_available(self) -> str | None:
        current_version = importlib_metadata.version(package)
        async with httpx.AsyncClient() as client:
            response = await client.get(f"https://pypi.org/pypi/{package}/json")

        latest_version = response.json()["info"]["version"]

        if current_version != latest_version:
            return latest_version

    @work(name="version_check")
    async def version_check(self) -> None:
        if new_version := await self.new_version_available():
            self.call_after_refresh(
                self.notify, f"New version {new_version} available", timeout=5
            )

    async def valid_password(self, app_data: InitAppResponse) -> bool:
        return await app_data.last_used_wallet.db.validate_key()

    async def encryption_password_callback(
        self, app_data: InitAppResponse, result: tuple[str, bool]
    ) -> None:
        password, store_in_keychain = result

        hashed = hashlib.sha256(bytes(password, "utf8")).hexdigest()
        app_data.last_used_wallet.db.set_encrypted_key(hashed)

        if not await self.valid_password(app_data):
            callback = partial(self.encryption_password_callback, app_data)
            self.push_screen(EncryptionPassword(message="Invalid Password"), callback)
            return

        if store_in_keychain:
            keyring.set_password("fluxwallet", "fluxwallet_user", hashed)

        self.install_screen(
            WalletLanding(
                app_data.last_used_wallet, app_data.networks, app_data.wallets
            ),
            name="wallet_landing",
        )
        self.push_screen("wallet_landing")

    async def on_mount(self) -> None:
        self.version_check()

        # push loading screen first, with logo

        app_data = await init_app()

        if not app_data.last_used_wallet:
            self.push_screen("welcome")
            return

        if app_data.encrypted_db:
            password_hash = keyring.get_password("fluxwallet", "fluxwallet_user")
            # decrypt main key and check for xprv
            if not password_hash:
                callback = partial(self.encryption_password_callback, app_data)
                self.push_screen(EncryptionPassword(), callback)
                return
            else:
                # this only ever needs to get set once
                app_data.last_used_wallet.db.set_encrypted_key(password_hash)

                if not await self.valid_password(app_data):
                    keyring.delete_password("fluxwallet", "fluxwallet_user")

                    callback = partial(self.encryption_password_callback, app_data)
                    self.push_screen(
                        EncryptionPassword(message="Invalid Password"), callback
                    )
                    return

        self.install_screen(
            WalletLanding(
                app_data.last_used_wallet, app_data.networks, app_data.wallets
            ),
            name="wallet_landing",
        )
        self.push_screen("wallet_landing")

    # fix these
    @on(MnemonicOverlay.WalletCreated)
    async def on_mnemonic_overlay_wallet_created(
        self, event: MnemonicOverlay.WalletCreated
    ):
        wallet_landing: WalletLanding = self.get_screen("wallet_landing")
        await wallet_landing.new_wallet_created(event.wallet)

    @on(ImportFromMnemonic.WalletCreated)
    async def on_import_from_mnemonic_wallet_created(
        self, event: ImportFromMnemonic.WalletCreated
    ):
        wallet_landing: WalletLanding = self.get_screen("wallet_landing")
        await wallet_landing.new_wallet_created(event.wallet)


# for console script
def run():
    # ubuntu by default doesn't have this set, so the colors are weird
    os.environ["COLORTERM"] = "truecolor"

    app = FluxWallet()
    app.run()


if __name__ == "__main__":
    run()



from functools import partial

from fluxwallet.mnemonic import Mnemonic
from textual.app import ComposeResult
from textual.containers import Grid, Horizontal
from textual.message import Message
from textual.reactive import var
from textual.screen import Screen
from textual.validation import Length
from textual.widget import Widget
from textual.widgets import Button, Input, Label, Select, Static

from lucidwallet.helpers import init_app
from lucidwallet.screens import MnemonicOverlay

languages = [
    "english",
    "spanish",
    "italian",
    "dutch",
    "french",
    "japanese",
    "chinese_simplified",
    "chinese_traditional",
]


# class Notification(Static):
#     def on_mount(self) -> None:
#         self.set_timer(3, self.remove)

#     def on_click(self) -> None:
#         self.remove()


class LanguagePicker(Widget):
    def __init__(self, language: str = "english"):
        self.language = language
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label("Select Language")
        yield Select(
            [(x, x) for x in languages],
            prompt="Select language",
            value=self.language,
        )

    def reset(self) -> None:
        self.language = "english"
        select = self.query_one("Select", Select)
        select.value = "english"

    def on_select_changed(self, event: Select.Changed):
        self.language = event.value


class CreateWallet(Screen):
    BINDINGS = [
        (
            "escape",
            "app.pop_screen()",
            "home",
        ),
    ]

    nickname = var("")

    # class WalletLandingRequested(Message):
    #     def __init__(self, wallet: str, wallets: list[str], networks: list[str]):
    #         self.wallet = wallet
    #         self.wallets = wallets
    #         self.networks = networks
    #         super().__init__()

    def compose(self):
        yield Static("Create Wallet", id="create_wallet_title")
        yield LanguagePicker()
        yield Input(
            "",
            placeholder="Wallet nickname",
            id="nickname",
            validators=[Length(maximum=15, minimum=2)],
        )
        yield Grid(id="mnemonic_grid")
        yield Horizontal(
            Button("Cancel", variant="warning", id="wallet_create_cancel"),
            Button("Create", id="wallet_create", disabled=True),
            id="create_wallet_button_container",
        )

    def on_mount(self) -> None:
        input = self.query_one("Input", Input)
        input.focus()

    def on_input_changed(self, event: Input.Changed) -> None:
        if event.validation_result.failures:
            # self.mount(
            #     Notification(event.validation_result.failures, variant="failure")
            # )
            return

        if event.value:
            self.nickname = event.value

    def reset_all(self) -> None:
        self.nickname = ""
        input = self.query_one("Input", Input)
        input.value = ""
        lang_picker = self.query_one("LanguagePicker", LanguagePicker)
        lang_picker.reset()
        input.focus()

    async def on_button_pressed(self, event: Button.Pressed):
        event.stop()
        if event.button.id == "wallet_create":
            print("WALLET CREATE")
            # wallet_name = self.query_one("Input", Input)
            app_data = await init_app()
            if self.nickname in app_data.wallets:
                self.notify("Wallet name exists")
                return

            lang_picker = self.query_one("LanguagePicker", LanguagePicker)
            mnemonic = Mnemonic(lang_picker.language).generate()

            self.app.push_screen(MnemonicOverlay(self.nickname, mnemonic))

            self.reset_all()
        else:
            self.reset_all()
            self.dismiss()

    def watch_nickname(self, new_value) -> None:
        create = self.query_one("#wallet_create", Button)

        if new_value:
            create.disabled = False
        else:
            create.disabled = True



from dataclasses import dataclass, field

from fluxwallet.db_new import Db, DbConfig, DbWallet
from fluxwallet.wallet import Wallet
from sqlalchemy import select
from textual.widgets import Static


class Notification(Static):
    def __init__(self, text: str, variant: str = "success", duration: int = 3):
        super().__init__(text)
        self.duration = duration
        if variant == "success":
            self.styles.outline = ("tall", "green")
        else:
            self.styles.outline = ("tall", "red")

    def on_mount(self) -> None:
        self.set_timer(self.duration, self.remove)

    def on_click(self) -> None:
        self.remove()


@dataclass
class InitAppResponse:
    last_used_wallet: Wallet | None = None
    wallets: list[str] = field(default_factory=list)
    encrypted_db: bool = False
    networks: list[str] = field(default_factory=list)


async def get_db_info() -> InitAppResponse:
    db = await Db.start()

    async with db as session:
        last_used_wallet: str | None = None
        res = await session.scalars(select(DbWallet).order_by(DbWallet.id))
        wallets = res.all()

        if wallets:
            res = await session.scalars(
                select(DbConfig.value).filter_by(variable="last_used_wallet")
            )
            last_used_wallet = res.first()

            if not last_used_wallet:
                last_used_wallet = wallets[0].name
                await session.merge(
                    DbConfig(variable="last_used_wallet", value=last_used_wallet)
                )
                await session.commit()

            if last_used_wallet:
                last_used_wallet = await Wallet.open(last_used_wallet)

            wallets = [x.name for x in wallets]
        else:
            wallets = []

    return InitAppResponse(last_used_wallet, wallets, db.encrypted)


async def init_app() -> InitAppResponse:
    res = await get_db_info()

    if res.last_used_wallet:
        # this coudl error
        keys = await res.last_used_wallet.keys_networks()
        res.networks = [x.network.name for x in keys]

    return res
